//=========================================================================
// Name:            IPipelineStep.h
// Purpose:         Describes a step in the audio pipeline.
//
// Authors:         Mooneer Salem
// License:
//
//  All rights reserved.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License version 2.1,
//  as published by the Free Software Foundation.  This program is
//  distributed in the hope that it will be useful, but WITHOUT ANY
//  WARRANTY; without even the implied warranty of MERCHANTABILITY or
//  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
//  License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, see <http://www.gnu.org/licenses/>.
//
//=========================================================================

#ifndef AUDIO_PIPELINE__I_PIPELINE_STEP_H
#define AUDIO_PIPELINE__I_PIPELINE_STEP_H

#include <memory>

class IPipelineStep
{
public:
    virtual ~IPipelineStep();
    
    // Returns required input sample rate.
    virtual int getInputSampleRate() const = 0;
    
    // Returns output sample rate after performing the pipeline step.
    virtual int getOutputSampleRate() const = 0;
    
    // Executes pipeline step.
    // Required parameters:
    //     inputSamples: Array of int16 values corresponding to input audio.
    //     numInputSamples: Number of samples in the input array.
    //     numOutputSamples: Location to store number of output samples.
    // Returns: Array of int16 values corresponding to result audio.
    virtual std::shared_ptr<short> execute(std::shared_ptr<short> inputSamples, int numInputSamples, int* numOutputSamples) = 0;
    
    // Resets internal state of the pipeline step.
    virtual void reset() { /* empty */ }
    
    // Operator overrides to ensure we use real-time O(1) allocator
    void* operator new(std::size_t sz);
    void* operator new[](std::size_t sz);
    void operator delete(void* ptr) noexcept;
    void operator delete(void* ptr, std::size_t size) noexcept;
    void operator delete[](void* ptr) noexcept;
    void operator delete[](void* ptr, std::size_t size) noexcept;
    
protected:
    template<typename T>
    static T* AllocRealtime_(int numElements) noexcept;
    
    template<typename T>
    static void FreeRealtime_(T* ptr) noexcept;
    
    template<typename T>
    struct RealtimeDeleter
    {
        void operator()(T* ptr) const
        {
            FreeRealtime_(ptr);
        }
    };
    
    template<typename T>
    struct RealtimeAllocator
    {
        typedef T value_type;
        RealtimeAllocator() noexcept { }
    
        template<class U> RealtimeAllocator(const RealtimeAllocator<U>&) noexcept {}
        template<class U> bool operator==(const RealtimeAllocator<U>&) const noexcept
        {
            return true;
        }
        template<class U> bool operator!=(const RealtimeAllocator<U>&) const noexcept
        {
            return false;
        }
    
        T* allocate(const size_t n) const noexcept
        {
            return AllocRealtime_<T>(n);
        }
    
        void deallocate(T* const p, size_t) const noexcept
        {
            FreeRealtime_(p);
        }
    };
    
private:
    static void* AllocRealtimeGeneric_(int size) noexcept;
    static void FreeRealtimeGeneric_(void* ptr) noexcept;
};

template<typename T>
T* IPipelineStep::AllocRealtime_(int numElements) noexcept
{
    return (T*)AllocRealtimeGeneric_(sizeof(T) * numElements);
}

template<typename T>
void IPipelineStep::FreeRealtime_(T* ptr) noexcept
{
    FreeRealtimeGeneric_(ptr);
}

#endif // AUDIO_PIPELINE__I_PIPELINE_STEP_H